/* transport.hpp                           
 *
 * Copyright (C) 2006 Sawtooth Consulting Ltd.
 *
 * This file is part of yaSSH.
 *
 * yaSSH is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * yaSSH is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */


/* This header defines SSH transport layer protocol types see RFC 4253
 */


#ifndef yaSSH_TRANSPORT_HPP
#define yaSSH_TRANSPORT_HPP

#include "yassh_types.hpp"
#include "socket_wrapper.hpp"
#include "random.hpp"
#include "sha.hpp"

namespace yaSSH {


static const char* Version = "SSH-2.0-yaSSH_0.0.1\r\n";


using TaoCrypt::RandomNumberGenerator;
class output_buffer;
class input_buffer;


enum Misc {
    MAX_PADDING  = 255,
    MAX_MAC      =  20,
    KEX_INIT_MAX =  10,
    MIN_PADDING  =   8,
    VERSION_SZ   = 128,
    DH_GRP1_SZ   = 128,
    COOKIE_SZ    =  16
};


enum Errors {
    TCP_CONNECT_ERROR  = -101,
    TCP_SEND_ERROR     = -102,
    GET_KEX_INIT_ERROR = -103,
    GET_DH_REPLY_ERROR = -104
};


// see 12.
enum MessageNumbers {
    SSH_MSG_DISCONNECT      =  1,
    SSH_MSG_IGNORE          =  2,
    SSH_MSG_UNIMPLEMENTED   =  3,
    SSH_MSG_DEBUG           =  4,
    SSH_MSG_SERVICE_REQUEST =  5,
    SSH_MSG_SERVICE_ACCEPT  =  6,
    SSH_MSG_KEXINIT         = 20,
    SSH_MSG_NEWKEYS         = 21,
    SSH_MSG_KEXDH_INIT      = 30,
    SSH_MSG_KEXDH_REPLY     = 31
};


// see 11.1.
enum DisconnectNumbers {
    SSH_DISCONNECT_HOST_NOT_ALLOWED_TO_CONNECT    =  1,
    SSH_DISCONNECT_PROTOCOL_ERROR                 =  2,
    SSH_DISCONNECT_KEY_EXCHANGE_FAILED            =  3,
    SSH_DISCONNECT_RESERVED                       =  4,
    SSH_DISCONNECT_MAC_ERROR                      =  5,
    SSH_DISCONNECT_COMPRESSION_ERROR              =  6,
    SSH_DISCONNECT_SERVICE_NOT_AVAILABLE          =  7,
    SSH_DISCONNECT_PROTOCOL_VERSION_NOT_SUPPORTED =  8,
    SSH_DISCONNECT_HOST_KEY_NOT_VERIFIABLE        =  9,
    SSH_DISCONNECT_CONNECTION_LOST                = 10,
    SSH_DISCONNECT_BY_APPLICATION                 = 11,
    SSH_DISCONNECT_TOO_MANY_CONNECTIONS           = 12,
    SSH_DISCONNECT_AUTH_CANCELLED_BY_USER         = 13,
    SSH_DISCONNECT_NO_MORE_AUTH_METHODS_AVAILABLE = 14,
    SSH_DISCONNECT_ILLEGAL_USER_NAME              = 15
};


// see 6.
struct packet {
    word32 packet_length;                 // total - self - mac
    byte   padding_length;                // length of random bytes
    byte*  payload;                       // packet_length - padding_length - 1
    byte   random_padding[MAX_PADDING];   // pad to at least block (min 8)
    byte   mac[MAX_MAC];                  // packet MAC

    packet() : payload(0) {}
    ~packet() { delete[] payload; }
};


struct nameList {
    word32 size;          // total size of list
    char*  list;          // the list

    nameList() : size(0), list(0) {}
    ~nameList();

    void SetList(const char*);
    void SetList(input_buffer&);
};


// Algorithm Negotiation see 7.1.
struct keyExchange {
    byte     msg;                           // SSH_MSG_KEXINIT
    byte     cookie[COOKIE_SZ];             // random bytes
    nameList kex_alogorithms;               // see supported ones above
    nameList server_host_key_alogorithms;
    nameList ciphers_client;
    nameList ciphers_server;
    nameList mac_client;
    nameList mac_server;
    nameList comp_client;
    nameList comp_server;
    nameList lang_client;
    nameList lang_server;
    boolean  first_kex_packet_follows;      // usually use
    word32   reserved;                      // for future extensions

    keyExchange() : msg(SSH_MSG_KEXINIT), first_kex_packet_follows(false),
                    reserved(0) {}
};


class Transport {
    Socket                socket_;
    RandomNumberGenerator rng_;

    char                  clientVersion_[VERSION_SZ];
    char                  serverVersion_[VERSION_SZ];

    byte                  K_[DH_GRP1_SZ + 1];         // shared secret
    byte                  priv_[DH_GRP1_SZ + 1];      // dh private x
    byte                  pub_[DH_GRP1_SZ + 1];       // dh public key e
    byte                  otherPub_[DH_GRP1_SZ + 1];  // dh other f
    byte                  sig_[DH_GRP1_SZ + 1];       // signature of H_
    byte                  hostN_[DH_GRP1_SZ + 1];     // host key RSA n
    word32                nSz_;                // sizeof of host n
    word32                kSz_;                // sizeof K
    word32                pubSz_;              // sizeof e
    word32                otherPubSz_;         // sizeof f
    word32                sigSz_;              // sizeof sig_
    word32                hostE_;              // host key RSA e

    TaoCrypt::SHA         sha_;        // for H_ = V_C || V_S || I_C || I_S ||
    byte                  H_[TaoCrypt::SHA::DIGEST_SIZE]; // K_S || e || f || K
public:
    int Connect(const char* host, word16 port = 22);
private:
    int  SendVersion();
    int  SendKexInit();
    int  SendKexDh();
    int  SendNewKeys();
    int  GetVersion();
    int  GetKexInit();
    int  GetDhReply();
    int  GetHostKey(input_buffer&);
    int  GetF(input_buffer&);
    int  GetSignature(input_buffer&);
    int  GetKexPacket(input_buffer&);
    int  GetKexDhReply(input_buffer&);
    int  VerifyKex();
    int  SetKeys();
    void WriteNameList(output_buffer&, nameList&);
};



}      // namespace
#endif // yaSSH_TRANSPORT_HPP

