/* socket_wrapper.cpp                           
 *
 * Copyright (C) 2006 Sawtooth Consulting Ltd.
 *
 * This file is part of yaSSH.
 *
 * yaSSH is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * yaSSH is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */


/* The socket wrapper source implements a Socket class that hides the 
 * differences between Berkely style sockets and Windows sockets, allowing 
 * transparent TCP access.
 */


#include "socket_wrapper.hpp"
#include <ctype.h>

#ifndef _WIN32
    #include <errno.h>
    #include <netdb.h>
    #include <unistd.h>
    #include <arpa/inet.h>
    #include <netinet/in.h>
    #include <sys/ioctl.h>
    #include <string.h>
#endif // _WIN32

#if defined(__sun) || defined(__SCO_VERSION__)
    #include <sys/filio.h>
#endif

#ifdef _WIN32
    const int SOCKET_EINVAL = WSAEINVAL;
    const int SOCKET_EWOULDBLOCK = WSAEWOULDBLOCK;
    const int SOCKET_EAGAIN = WSAEWOULDBLOCK;
#else
    const int SOCKET_EINVAL = EINVAL;
    const int SOCKET_EWOULDBLOCK = EWOULDBLOCK;
    const int SOCKET_EAGAIN = EAGAIN;
#endif // _WIN32


namespace yaSSH {


Socket::Socket(socket_t s) 
    : socket_(s), wouldBlock_(false)
{}


Socket::~Socket()
{
    CloseSocket();
}


void Socket::CloseSocket()
{
    if (socket_ != INVALID_SOCKET) {
#ifdef _WIN32
        closesocket(socket_);
#else
        close(socket_);
#endif
        socket_ = INVALID_SOCKET;
    }
}


uint Socket::GetReady() const
{
#ifdef _WIN32
    unsigned long ready = 0;
    ioctlsocket(socket_, FIONREAD, &ready);
#else
    /*
       64-bit Solaris requires the variable passed to
       FIONREAD be a 32-bit value.
    */
    unsigned int ready = 0;
    ioctl(socket_, FIONREAD, &ready);
#endif

    return ready;
}


uint Socket::Send(const byte* buf, unsigned int sz, int flags) const
{
    const byte* pos = buf;
    const byte* end = pos + sz;

    assert(socket_ != INVALID_SOCKET);

    while (pos != end) {
        int sent = ::send(socket_, reinterpret_cast<const char *>(pos),
                          static_cast<int>(end - pos), flags);

        if (sent == -1)
            return 0;

        pos += sent;
    }

    return sz;
}


uint Socket::Receive(byte* buf, unsigned int sz, int flags)
{
    assert(socket_ != INVALID_SOCKET);
    wouldBlock_ = false;

    int recvd = ::recv(socket_, reinterpret_cast<char *>(buf), sz, flags);

    // idea to seperate error from would block by arnetheduck@gmail.com
    if (recvd == -1) {
        if (GetLastError() == SOCKET_EWOULDBLOCK || 
            GetLastError() == SOCKET_EAGAIN) {
            wouldBlock_ = true;
            return 0;
        }
    }
    else if (recvd == 0)
        return static_cast<uint>(-1);

    return recvd;
}


// wait if blocking for input, return false for error
bool Socket::Wait()
{
    byte b;
    return Receive(&b, 1, MSG_PEEK) != static_cast<uint>(-1);
}


void Socket::ShutDown(int how)
{
    assert(socket_ != INVALID_SOCKET);
    shutdown(socket_, how);
}


int Socket::GetLastError()
{
#ifdef _WIN32
    return WSAGetLastError();
#else
    return errno;
#endif
}


bool Socket::WouldBlock() const
{
    return wouldBlock_;
}


void TcpSocket(socket_t& sockfd, struct sockaddr_in& addr, const char* peer,
                word16 port)
{
    const char* host = peer;

    // peer could be in human readable form
    if (isalpha(peer[0])) {
        struct hostent* entry = gethostbyname(peer);

        if (entry) {
            struct sockaddr_in tmp;
            memset(&tmp, 0, sizeof(struct sockaddr_in));
            memcpy(&tmp.sin_addr.s_addr, entry->h_addr_list[0],
                   entry->h_length);
            host = inet_ntoa(tmp.sin_addr);
        }
    }

    memset(&addr, 0, sizeof(struct sockaddr_in));
    sockfd = socket(AF_INET, SOCK_STREAM, 0);

    addr.sin_family = AF_INET;
    addr.sin_port = htons(port);
    addr.sin_addr.s_addr = inet_addr(host);
}


void StartTCP()
{
#ifdef _WIN32
    WSADATA wsd; 
    WSAStartup(0x0002, &wsd);
#endif
}


uint Socket::Connect(const char* peer, word16 port)
{
    StartTCP();

    struct sockaddr_in addr;
    TcpSocket(socket_, addr, peer, port);

    if (connect(socket_, (const struct sockaddr*)&addr, sizeof(addr)) != 0)
        return 1;

    return 0;
}


} // namespace
