/* transport.cpp                           
 *
 * Copyright (C) 2006 Sawtooth Consulting Ltd.
 *
 * This file is part of yaSSH.
 *
 * yaSSH is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * yaSSH is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */


/* This file implements SSH transport layer protocol see RFC 4253
 */


#include "transport.hpp"
#include "buffer.hpp"
#include "dh.hpp"
#include "rsa.hpp"
#include "asn.hpp"
#include <string.h>


namespace yaSSH {


word32 htoni(word32 host)
{
#ifdef LITTLE_ENDIAN_ORDER
    return TaoCrypt::ByteReverse(host);
#endif
    return host;
}

word32 ntohi(word32 network)
{
#ifdef LITTLE_ENDIAN_ORDER
    return TaoCrypt::ByteReverse(network);
#endif
    return network;
}


nameList::~nameList()
{
    delete[] list;
}


void nameList::SetList(const char* str)
{
    size = strlen(str);

    if (size) {
        list = new char[size + 1];
        strncpy(list, str, size);
    }
}


void nameList::SetList(input_buffer& buffer)
{
    buffer.read(&size, sizeof(size));
    size = ntohi(size);

    if (size) {
        list = new char[size + 1];
        buffer.read(list, size);
        list[size] = 0;
    }
}


int Transport::Connect(const char* host, word16 port)
{
    if (socket_.Connect(host, port) != 0)
        return TCP_CONNECT_ERROR;

    if (GetVersion() != 0)
        return -1;

    if (SendVersion() != 0)
        return TCP_SEND_ERROR;

    if (SendKexInit() != 0)
        return TCP_SEND_ERROR;

    if (GetKexInit() != 0)
        return GET_KEX_INIT_ERROR;

    if (SendKexDh() != 0)
        return TCP_SEND_ERROR;

    if (GetDhReply() != 0)
        return GET_DH_REPLY_ERROR;

    if (SendNewKeys() != 0)
        return TCP_SEND_ERROR;

    return 0;
}


namespace {

const char* kexMethods = "diffie-hellman-group1-sha1";
const char* kexPk      = "ssh-rsa";
const char* kexCipher  = "3des-cbc";
const char* kexMac     = "hmac-sha1";
const char* kexComp    = "none";
const char* kexLang    = "";


const byte group1_p[] = 
{
    0xFF, 0xFF, 0xFF, 0xFF,
    0xFF, 0xFF, 0xFF, 0xFF,
    0xC9, 0x0F, 0xDA, 0xA2,
    0x21, 0x68, 0xC2, 0x34,
    0xC4, 0xC6, 0x62, 0x8B,
    0x80, 0xDC, 0x1C, 0xD1,

    0x29, 0x02, 0x4E, 0x08,
    0x8A, 0x67, 0xCC, 0x74,
    0x02, 0x0B, 0xBE, 0xA6,
    0x3B, 0x13, 0x9B, 0x22,
    0x51, 0x4A, 0x08, 0x79,
    0x8E, 0x34, 0x04, 0xDD,

    0xEF, 0x95, 0x19, 0xB3,
    0xCD, 0x3A, 0x43, 0x1B,
    0x30, 0x2B, 0x0A, 0x6D,
    0xF2, 0x5F, 0x14, 0x37,
    0x4F, 0xE1, 0x35, 0x6D,
    0x6D, 0x51, 0xC2, 0x45,

    0xE4, 0x85, 0xB5, 0x76,
    0x62, 0x5E, 0x7E, 0xC6,
    0xF4, 0x4C, 0x42, 0xE9,
    0xA6, 0x37, 0xED, 0x6B,
    0x0B, 0xFF, 0x5C, 0xB6,
    0xF4, 0x06, 0xB7, 0xED,

    0xEE, 0x38, 0x6B, 0xFB,
    0x5A, 0x89, 0x9F, 0xA5,
    0xAE, 0x9F, 0x24, 0x11,
    0x7C, 0x4B, 0x1F, 0xE6,
    0x49, 0x28, 0x66, 0x51,
    0xEC, 0xE6, 0x53, 0x81,

    0xFF, 0xFF, 0xFF, 0xFF,
    0xFF, 0xFF, 0xFF, 0xFF
};


const byte group1_g[] = { 0x02 };


// build for sending
void BuildKeyExchange(keyExchange& ex, RandomNumberGenerator& rng)
{
    rng.GenerateBlock(ex.cookie, COOKIE_SZ);
    ex.kex_alogorithms.SetList(kexMethods);
    ex.server_host_key_alogorithms.SetList(kexPk);
    ex.ciphers_client.SetList(kexCipher);
    ex.ciphers_server.SetList(kexCipher);
    ex.mac_client.SetList(kexMac);
    ex.mac_server.SetList(kexMac);
    ex.comp_client.SetList(kexComp);
    ex.comp_server.SetList(kexComp);
}


word32 GetKeyExchangeSize(keyExchange& ex)
{
    return sizeof(ex.msg) + sizeof(ex.cookie) +
        ex.kex_alogorithms.size + ex.server_host_key_alogorithms.size +
        ex.ciphers_client.size + ex.ciphers_server.size + ex.mac_client.size +
        ex.mac_server.size + ex.comp_client.size + ex.comp_server.size +
        10 * sizeof(word32) /* each nameList length */ +
        sizeof(ex.first_kex_packet_follows) + sizeof(ex.reserved);
}


void WriteNameList(output_buffer& buffer, nameList& name)
{
    word32 networkSize = htoni(name.size);

    buffer.write(&networkSize, sizeof(networkSize));
    buffer.write(name.list, name.size);
}


void WriteKeyExchangeToBuffer(keyExchange& ex, output_buffer& buffer)
{
    buffer.write(&ex, sizeof(ex.msg) + sizeof(ex.cookie));
    
    WriteNameList(buffer, ex.kex_alogorithms);
    WriteNameList(buffer, ex.server_host_key_alogorithms);
    WriteNameList(buffer, ex.ciphers_client);
    WriteNameList(buffer, ex.ciphers_server);
    WriteNameList(buffer, ex.mac_client);
    WriteNameList(buffer, ex.mac_server);
    WriteNameList(buffer, ex.comp_client);
    WriteNameList(buffer, ex.comp_server);
    WriteNameList(buffer, ex.lang_client);
    WriteNameList(buffer, ex.lang_server);

    buffer.write(&ex.first_kex_packet_follows,
                 sizeof(ex.first_kex_packet_follows));
    buffer.write(&ex.reserved, sizeof(ex.reserved));
}


void WriteKexDhToBuffer(const byte* pub, word32 length, output_buffer& buffer)
{
    buffer[AUTO] = SSH_MSG_KEXDH_INIT;
    word32 size = htoni(length);
    buffer.write(&size, sizeof(size));
    buffer.write(pub, length);

}


void WritePacketStart(word32 size, word32 padLen, output_buffer& buffer)
{
    word32 networkSz = htoni(size - sizeof(word32));  // don't include self
    buffer.write(&networkSz, sizeof(size));
    buffer[AUTO] = padLen;
}


void WritePacketEnd(word32 padLen, output_buffer& buffer)
{
    buffer.write(padLen);
}


word32 GetSize(input_buffer& buffer)
{
    word32 ret;
    buffer.read(&ret, sizeof(ret));

    return ntohi(ret);
}


word32 GetPacketSize(word32 size, word32 padding, word32& padLen)
{
    word32 ret = sizeof(word32) + 1;  // packet length + pad length byte
    padLen = MIN_PADDING + padding - ((size + ret) % padding);

    return ret + padLen;
}


static byte buffer[4096];
static word32 idx = 0;

// UPdate with prepended size in network order, then data
void UpdateHash(TaoCrypt::SHA& sha, const byte* data, word32 size)
{
    word32 netSz = htoni(size);
    byte   tmp[4];

    memcpy(tmp, &netSz, sizeof(tmp));

    sha.Update(tmp, sizeof(tmp));
    sha.Update(data, size);

    memcpy(&buffer[0] + idx, tmp, sizeof(tmp));
    idx += sizeof(tmp);
    memcpy(&buffer[0] + idx, data, size);
    idx += size;
}


} // anonymous namespace 


int Transport::SendKexInit()
{
    keyExchange ex;

    BuildKeyExchange(ex, rng_);

    word32 padLen;
    word32 size  = GetKeyExchangeSize(ex);
    word32 kexSz = size;
    size += GetPacketSize(size, MIN_PADDING, padLen);

    output_buffer buffer(size);
    WritePacketStart(size, padLen, buffer);
    word32 kexBegin = buffer.get_size();
    WriteKeyExchangeToBuffer(ex, buffer);
    WritePacketEnd(padLen, buffer);

    // Update H_ (client kexInit)
    UpdateHash(sha_, buffer.get_buffer() + kexBegin, kexSz);

    if (socket_.Send(buffer.get_buffer(), buffer.get_size()) !=
                     buffer.get_size())
        return -1;

    return 0;
}


int Transport::SendVersion()
{
    uint sz = strlen(Version);

    if (socket_.Send(reinterpret_cast<const byte*>(Version), sz) != sz)
        return -1;
    return 0;
}


int Transport::GetVersion()
{
    socket_.Wait();
    word32 ready = socket_.GetReady();

    if (!ready) return -1;

    input_buffer buffer(ready);
    socket_.Receive(buffer.get_buffer(), ready);
    buffer.add_size(ready);

    byte*  begin    = buffer.get_buffer();
    word32 startIdx = buffer.get_current();

    char protocol[3];
    buffer.read(protocol, sizeof(protocol));

    if (strncmp(protocol, "SSH", sizeof(protocol)))
        return -1;

    byte tmp;
    while( (tmp = buffer[AUTO]) != '\n')
        ;  // continue

    // TODO: if \r before subtract from server version update

    word32 endIdx = buffer.get_current();

    // Update H_ (client then server version)
    UpdateHash(sha_, (byte*)Version, strlen(Version) - 2);  // remove \r\n
    UpdateHash(sha_, begin, endIdx - startIdx - 1);
    
    return 0;
}


int Transport::GetKexPacket(input_buffer& buffer)
{
    word32 size = GetSize(buffer);
    byte   padLength = buffer[AUTO];

    if (padLength < 4)  // initial set to 4
        return -1;

    // Update H_ (server kexInit)
    UpdateHash(sha_, buffer.get_buffer() + buffer.get_current(),
               size - 1 - padLength);

    keyExchange ex;

    ex.msg = buffer[AUTO];
    if (ex.msg != SSH_MSG_KEXINIT)
        return -1;
    buffer.read(ex.cookie, COOKIE_SZ);
    
    ex.kex_alogorithms.SetList(buffer);
    ex.server_host_key_alogorithms.SetList(buffer);
    ex.ciphers_client.SetList(buffer);
    ex.ciphers_server.SetList(buffer);
    ex.mac_client.SetList(buffer);
    ex.mac_server.SetList(buffer);
    ex.comp_client.SetList(buffer);
    ex.comp_server.SetList(buffer);
    ex.lang_client.SetList(buffer);
    ex.lang_server.SetList(buffer);

    buffer.read(&ex.first_kex_packet_follows,
                sizeof(ex.first_kex_packet_follows));
    buffer.read(&ex.reserved, sizeof(ex.reserved));

    while (padLength--) byte tmp = buffer[AUTO];

    return 0;
}


int Transport::GetF(input_buffer& buffer)
{
    otherPubSz_ = GetSize(buffer);
    buffer.read(otherPub_, otherPubSz_);

    // Update H_ (e)
    UpdateHash(sha_, pub_, pubSz_);

    // Update H_ (f)
    UpdateHash(sha_, otherPub_, otherPubSz_);

    return 0;
}


int Transport::GetHostKey(input_buffer& buffer)
{
    word32 size   = GetSize(buffer);
    // Update H_ (server host key)
    UpdateHash(sha_, buffer.get_buffer() + buffer.get_current(), size);
    word32 typeSz = GetSize(buffer);

    char type[32];
    if (typeSz > sizeof(type)) return -1;
    buffer.read(type, typeSz);

    if (strncmp("ssh-rsa", type, 7) != 0)
        return -1;

    word32 eSz = GetSize(buffer);
    assert(eSz == 1);
    hostE_ = buffer[AUTO];
    
    nSz_ = GetSize(buffer);
    assert(nSz_ <= sizeof(hostN_));
    buffer.read(hostN_, nSz_);

    return 0;
}


int Transport::GetSignature(input_buffer& buffer)
{
    word32 size = GetSize(buffer);
    word32 typeSz = GetSize(buffer);

    char type[32];
    if (typeSz > sizeof(type)) return -1;
    buffer.read(type, typeSz);

    if (strncmp("ssh-rsa", type, 7) != 0)
        return -1;

    sigSz_ = GetSize(buffer);
    assert(sigSz_ <= sizeof(sig_));
    buffer.read(sig_, sigSz_);

    return 0;
}


int Transport::GetKexDhReply(input_buffer& buffer)
{
    word32 size = GetSize(buffer);
    byte   padLength = buffer[AUTO];

    if (padLength < MIN_PADDING)
        return -1;

    byte msg = buffer[AUTO];
    if (msg != SSH_MSG_KEXDH_REPLY)
        return -1;

    if (GetHostKey(buffer) != 0)
        return -1;

    if (GetF(buffer) != 0)
        return -1;

    if (GetSignature(buffer) != 0)
        return -1;

    while (padLength--) byte tmp = buffer[AUTO];

    return 0;
}


int Transport::GetKexInit()
{
    socket_.Wait();
    word32 ready = socket_.GetReady();

    if (!ready) return -1;

    input_buffer buffer(ready);
    socket_.Receive(buffer.get_buffer(), ready);
    buffer.add_size(ready);

    if (GetKexPacket(buffer)) return -1;

    return 0;
}


int Transport::GetDhReply()
{
    socket_.Wait();
    word32 ready = socket_.GetReady();

    if (!ready) return -1;

    input_buffer buffer(ready);
    socket_.Receive(buffer.get_buffer(), ready);
    buffer.add_size(ready);

    if (GetKexDhReply(buffer)) return -1;
    if (VerifyKex() != 0) return -1;
    if (SetKeys() != 0) return -1;

    return 0;
}


int Transport::VerifyKex()
{
    // get shared secret
    TaoCrypt::Integer p(group1_p, sizeof(group1_p));
    TaoCrypt::Integer g(group1_g, sizeof(group1_g));

    TaoCrypt::DH dh(p, g);
    dh.Agree(K_, priv_, otherPub_, otherPubSz_);

    kSz_ = DH_GRP1_SZ;
    if (K_[0] & 0x80) {  // add leading 0 per standard
        byte tmp[DH_GRP1_SZ];
        memcpy(tmp, K_, sizeof(tmp));
        K_[0] = 0;
        memcpy(K_ + 1, tmp, sizeof(tmp));
        kSz_ += 1;
    }

    // final H_ (K)
    UpdateHash(sha_, K_, kSz_);
    sha_.Final(H_);

    // verify signature
    TaoCrypt::RSA_PublicKey key;
    key.Initialize(TaoCrypt::Integer(hostN_, nSz_), hostE_);

    TaoCrypt::RSAES_Encryptor verify(key);

    TaoCrypt::Source build;
    byte second[TaoCrypt::SHA::DIGEST_SIZE];  // second hash?
    TaoCrypt::SHA sha;
    sha.Update(H_, sizeof(H_));
    sha.Final(second);
    TaoCrypt::Signature_Encoder(second, sizeof(second), TaoCrypt::SHAh, build);

    bool answer = verify.SSL_Verify(build.get_buffer(), build.size(), sig_);
    assert(answer);

    return 0;
}


int Transport::SendKexDh()
{
    TaoCrypt::Integer p(group1_p, sizeof(group1_p));
    TaoCrypt::Integer g(group1_g, sizeof(group1_g));

    TaoCrypt::DH dh(p, g);
    dh.GenerateKeyPair(rng_, priv_, pub_);

    pubSz_ = DH_GRP1_SZ;
    if (pub_[0] & 0x80) {  // add leading 0 per standard
        byte tmp[DH_GRP1_SZ];
        memcpy(tmp, pub_, sizeof(tmp));
        pub_[0] = 0;
        memcpy(pub_ + 1, tmp, sizeof(tmp));
        pubSz_ += 1;
    }

    word32 padLen;
    word32 size = pubSz_ + 1 + sizeof(word32);  // + msg id + size of e
    size += GetPacketSize(size, MIN_PADDING, padLen);

    output_buffer buffer(size);
    WritePacketStart(size, padLen, buffer);
    WriteKexDhToBuffer(pub_, pubSz_, buffer);
    WritePacketEnd(padLen, buffer);

    if (socket_.Send(buffer.get_buffer(), buffer.get_size()) !=
                     buffer.get_size())
        return -1;

    return 0;
}


int Transport::SetKeys()
{
    TaoCrypt::SHA sha;

    byte k1[TaoCrypt::SHA::DIGEST_SIZE];
    byte k2[TaoCrypt::SHA::DIGEST_SIZE];

    byte X = 'A';
    byte tmp[4];   // for network k size

    // K1 = HASH(K || H || X || session_id)   (X is e.g., "A")
    // K encoded as mpint
    word32 netSz = htoni(kSz_);
    memcpy(tmp, &netSz, sizeof(tmp));

    sha.Update(tmp, sizeof(tmp));
    sha.Update(K_, kSz_);
    sha.Update(H_, TaoCrypt::SHA::DIGEST_SIZE);
    sha.Update(&X, 1);
    sha.Update(H_, TaoCrypt::SHA::DIGEST_SIZE);

    sha.Final(k1);

    // K2 = HASH(K || H || K1), K3 = HASH(K || H || K1 || K2)
    sha.Update(tmp, sizeof(tmp));
    sha.Update(K_, kSz_);
    sha.Update(H_, TaoCrypt::SHA::DIGEST_SIZE);
    sha.Update(k1, TaoCrypt::SHA::DIGEST_SIZE);
    sha.Final(k2);

    return 0;
}


int Transport::SendNewKeys()
{
    byte   msg  = SSH_MSG_NEWKEYS;
    word32 size = sizeof(msg);
    word32 padLen;

    size += GetPacketSize(size, MIN_PADDING, padLen);

    output_buffer buffer(size);
    WritePacketStart(size, padLen, buffer);
    buffer[AUTO] = msg;
    WritePacketEnd(padLen, buffer);

    if (socket_.Send(buffer.get_buffer(), buffer.get_size()) !=
                     buffer.get_size())
        return -1;

    return 0;
}



} // namespace
